/* alltoalltest.c
 * compile with:  mpicc ./alltoalltest.c -lm -o alltoalltest
 *
 * This program can be used to measure the throughput of 
 * all-to-all communication on a number of nodes 
 *
 * USAGE: mpirun -np <n> alltoalltest <t> <s>
 *   <n> test on n processors
 *   <t> perform t tests for each message size
 *   <s> largest message size to test is 2^s floats
 * EXAMPLE: mpirun -np 4 alltoalltest 25 19
 *
 *
 * Carsten Kutzner, 21.7.2006 
 * 
 */

#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>
#include "math.h"

/* Activate MPE features if you want to get logfiles to further
 * analyse the communication on each individual processor */
/* #define USE_MPE */
#ifdef USE_MPE
#include "mpe.h"
#define GMX_MPE_LOG(event) MPE_Log_event(event, 0, "")
#else
#define GMX_MPE_LOG(event)
#endif

/* Check the results of the all-to-alls */
/* #define CHECK */

/* Category: */
/* 0 1 2 3 4  5  6  7   8   9   10   11   12   13   14    15    16    17     18     19     */
/* Max number of floats to send to each process for this category: */
/* 1 2 4 8 16 32 64 128 256 512 1024 2048 4096 8192 16384 32768 65536 131072 262144 524288 */

#define CATEGORIES 20  /* max. message size is 2^19 floats (2 MB). */ 
                       /* MPE: put each message size in a color-coded category, so the 
                        * MPE graphics is clear laid out */
#define OFFSET 0       /* do not test the first OFFSET categories */

#define MASTER (procid==0)  /* master process (used for output) */


#ifdef CHECK
static void initialize_data(float *d, float *r, int elements, int procid)
/* initialize data with test numbers */
{
  int   i;
  float j = 0.0;
  
  j += 0.1;
  
  for (i = 0; i < elements; i++)
  {
    d[i] = j + i + procid / 10.0f;  /* data to send */
    r[i] = 0.0;                     /* zero out receive data */
  }
}

static void print_data(float *d, int elements, int procs, int procid, char *s)
/* prints the test data for each proc in an ordered fashion */
{
  int i,j;
  
  if (MASTER)
    fprintf(stdout,"\n%s\n",s);

  for (j=0; j<procs; j++)
  {
    if (procid == j)
    {
      fprintf(stdout,"proc %d (%d floats): ",procid,elements);
  
      for (i=0; i<elements; i++)
        fprintf(stdout,"%4.1f ",d[i]);
      fprintf(stdout,"\n");
    }
    MPI_Barrier(MPI_COMM_WORLD);
  }

  if (MASTER)
    fprintf(stdout,"\n");
  fflush(stdout);
}
#endif

/* Ordered all-to-all communication routine for single-processor
 * nodes. Ring-SendRecv scheme. 
 */
static int RSR_Alltoall(void *sendbuf, int sendcount, MPI_Datatype sendtype,
                        void *recvbuf, int recvcount, MPI_Datatype recvtype,
		        MPI_Comm comm)
{
   MPI_Status  status;
   MPI_Aint    sendextent, recvextent;
   int         sendchunk , recvchunk;
   int         source, dest;
   int         i, ncpu, cpuid;
   
   /* find out how many procs we have in our comm */
   MPI_Comm_size(comm,&ncpu);
   /* I am node number ... */
   MPI_Comm_rank(comm,&cpuid);
   /* Get extent of send and receive types */
   MPI_Type_extent(sendtype, &sendextent);
   MPI_Type_extent(recvtype, &recvextent);
   /* number of bytes of each message to send: */
   sendchunk=sendcount*sendextent;
   recvchunk=recvcount*recvextent;

   for (i=0; i<ncpu; i++)
   {
     /* send to dest = (cpuid + i) */
     dest = (cpuid + i) % ncpu;
     /* receive from source = (cpuid - i) */
     source   = (ncpu + cpuid - i) % ncpu;
          
     MPI_Sendrecv((char*)sendbuf + dest  *sendchunk, sendcount, sendtype, dest  , 0,
                  (char*)recvbuf + source*recvchunk, recvcount, recvtype, source, 0,
		  comm, &status);

     if ((sendchunk > 16384) && (i < ncpu-1)) 
       MPI_Barrier(comm);
   }
   return 0;
}


static int IIR_Alltoall(void *sendbuf, int sendcount, MPI_Datatype sendtype,
                        void *recvbuf, int recvcount, MPI_Datatype recvtype,
		        MPI_Comm comm, int procs_pn)
{
   MPI_Request sr_requests[2*procs_pn];
   MPI_Status  sr_statuses[2*procs_pn];
   MPI_Aint    sendextent, recvextent;
   int         sendchunk , recvchunk;
   int         destcpu   , sourcecpu ;
   int         destnode  , sourcenode;
   int         i,j;
   static int  ncpus, cpuid, nnodes, nodeid;
   int         sbarr; 
   /* choose sbarr = 4096 for dual-processor nodes, 
    * 16384 for single-processor nodes */ 
   if (procs_pn == 1)
     sbarr = 16384;
   else 
     sbarr = 4096;
   
   /* find out how many procs we have in our comm */
   MPI_Comm_size(comm,&ncpus);
   /* I am node number ... */
   MPI_Comm_rank(comm,&cpuid);
   /* number of nodes */
   nnodes = ncpus / procs_pn;
   /* my nodeid */
   nodeid = cpuid / procs_pn;
   /* Get extent of send and receive types */
   MPI_Type_extent(sendtype, &sendextent);
   MPI_Type_extent(recvtype, &recvextent);
   /* number of bytes of each message to send: */
   sendchunk=sendcount*sendextent;
   recvchunk=recvcount*recvextent;
   
   if ((ncpus % procs_pn) != 0)
   {
     fprintf(stderr, "\nERROR: # of CPUs is not divisible by # of nodes - use another algorithm!\n");
     exit(1);
   }

   /* loop over multi-processor nodes */
   for (i=0; i<nnodes; i++)
   {
     destnode   = (         nodeid + i) % nnodes;  /* send to destination node */
     sourcenode = (nnodes + nodeid - i) % nnodes;  /* receive from source node */
     for (j=0; j<procs_pn; j++)
     {
       sourcecpu = sourcenode*procs_pn + j; /* source of data */
       destcpu   = destnode  *procs_pn + j; /* destination of data */	 
       MPI_Irecv((char*)recvbuf + sourcecpu*recvchunk, recvcount, recvtype, sourcecpu, 0, comm, &sr_requests[j+procs_pn]);
       MPI_Isend((char*)sendbuf + destcpu  *sendchunk, sendcount, sendtype, destcpu  , 0, comm, &sr_requests[j         ]);
      }
     MPI_Waitall(2*procs_pn,sr_requests,sr_statuses);

     /* separate large messages by a barrier */
     if ((sendchunk > sbarr) && (i < nnodes-1))
       MPI_Barrier(comm);	
   }
   return 0;
}


static double do_test(int n, int ev_start, int ev_finish, int comm_type, 
                      int nprocs, int procs_per_node, int nnodes, int procid,
                      float *sdata, float *rdata)
{
  double t_start, t_finish;
	
#ifdef CHECK
  initialize_data(sdata, rdata, nprocs*n, procid);
  print_data(sdata, nprocs*n, nprocs, procid,"\nData initialized as:"); 
#endif	 
     
  if (MASTER)
  {	 
    fprintf(stdout,"sending %7d floats (%8d bytes) to %d processes (%d nodes) took ...",
      n, n*((int) sizeof(float)), nprocs, nnodes);
    fflush(stdout);			
  }
			
  MPI_Barrier(MPI_COMM_WORLD);
  GMX_MPE_LOG(ev_start);
  t_start = MPI_Wtime();

#define ARGS sdata,n,MPI_FLOAT,rdata,n,MPI_FLOAT,MPI_COMM_WORLD

  switch (comm_type)
  { 
    case 0:    
      MPI_Alltoall(ARGS);
      break;
    case 1:    
      RSR_Alltoall(ARGS);
      break;
    case 2:
      IIR_Alltoall(ARGS, procs_per_node);
      break;
    default:	
      fprintf(stderr,"ERROR: No comm_type %d.",comm_type);
      break;
    }
  MPI_Barrier(MPI_COMM_WORLD);
  GMX_MPE_LOG(ev_finish);
  t_finish = MPI_Wtime();
  
  if (MASTER) 
  {
    fprintf(stdout,"%10.5f seconds\n",t_finish-t_start);
    fflush(stdout);
  }

#ifdef CHECK
  print_data(rdata, nprocs*n, nprocs, procid,"Data after alltoall:"); 
#endif

  return t_finish-t_start;
}


int main(int argc, char *argv[])
{
  int        i,j;
  double     dummy, dum1;

  /* Variables describing the MPI universe: */
  int        nprocs;              /* number of CPUs                               */
  int        nnodes;              /* number of nodes                              */
  int        procs_per_node;      /* CPUs on each computer node (user input)      */  
  int        procid;              /* MPI rank of process                          */
  double     t_res;               /* MPI timer resolution [seconds]               */
  char       mpi_hostname[MPI_MAX_PROCESSOR_NAME];
  int        resultlen;           /* actual length of node name                   */
  char       *all_hostnames;      /* holds hostnames of each rank                 */

  /* MPE logging events */
  int        ev_start[CATEGORIES], ev_finish[CATEGORIES]; 
#ifdef USE_MPE			   
  char *color[20] = { "orange1 ", "orange2 ", "orange3 ", "orange4 ", \
                      "red1    ", "red2    ", "red3    ", "red4    ", \
                      "magenta1", "magenta2", "magenta3", "magenta4", \
                      "blue1   ", "blue2"   , "blue3"   , "blue4"   , \
                      "green1  ", "green2"  , "green3"  , "green4"  , \
                    };
  char desc[] = "123456789012";				   
#endif

  /* Here the data resides that has to be transferred: */
  float      *sdata, *rdata;      /* send data, receive data                      */

  /* Communication patterns: */	
  int        commtype;            /* Communication pattern: 
                                   * 0 = MPI_Alltoall 
                                   * 1 = Ring Sendrecv all-to-all
                                   * 2 = Isend/Irecv all-to-all                   */
  const char *typestring[3] = {"MPI", "RSR", "IIR"};

  /* For the analysis of the timings: */	
  int        tests;               /* number of tests to average over (user input) */
  int        howmanyfloats[CATEGORIES]; /* number of floats in a category (2^n)   */
  int        smax;                /* size of the larges messages to test in bytes */
  double     *t_comm[CATEGORIES]; /* individual time for each test                */
  double     t_aver[CATEGORIES];  /* average time of a number of tests            */
  double     t_min[CATEGORIES];   /* shortest time a test takes in a category     */
  double     t_max[CATEGORIES];   /* longest time a test takes in a category      */
  double     sdev[CATEGORIES];    /* standard deviation for each category         */
  int        cat_max;             /* max categories to actually test (user input) */
			
  /* output results in a file: */
  FILE       *f_ptr;
  char       filename[256];


  /* initialize parallel universe */
  (void) MPI_Init(&argc,&argv);
  (void) MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  (void) MPI_Comm_rank(MPI_COMM_WORLD, &procid);
  (void) MPI_Get_processor_name( mpi_hostname, &resultlen );
  t_res = MPI_Wtick();


  /* Has the user specified everything? */
  if (argc != 3) 
  {
    if (MASTER) 
    {
      fprintf(stderr, "USAGE: mpirun -np <n> alltoalltest <t> <s>\n");
      fprintf(stderr, "   <n> test on n processors\n");
      fprintf(stderr, "   <t> perform t tests for each message size\n");
      fprintf(stderr, "   <s> largest message size to test is 2^s floats\n");
      fprintf(stderr, "EXAMPLE: mpirun -np 4 alltoalltest 25 19\n");
    }			
    exit(0);
  }

  /* Determine number of CPUs that share a node's network interface: */
  procs_per_node = 0;
  all_hostnames = calloc(nprocs,MPI_MAX_PROCESSOR_NAME);
  MPI_Allgather(mpi_hostname , MPI_MAX_PROCESSOR_NAME, MPI_CHAR, 
                all_hostnames, MPI_MAX_PROCESSOR_NAME, MPI_CHAR, MPI_COMM_WORLD);
  for (i = 0; i < nprocs; i++)
  {
    if (strcmp( &all_hostnames[i*MPI_MAX_PROCESSOR_NAME], all_hostnames) == 0)
      procs_per_node++;
  }
  free(all_hostnames);

  /* Determine the number of nodes: */
  if (nprocs % procs_per_node)
  {
    if (MASTER) 
      fprintf(stderr,"ERROR: Number of processes (%d) is not divisible by procs_per_node (%d).\n",
        nprocs, procs_per_node);
    exit(0);
  }
  nnodes = nprocs / procs_per_node;

  /* Determine the number of tests to average over: */
  tests = atoi(argv[1]);
  if (tests < 1)
  {
    if (MASTER)
    {
      fprintf(stderr,"ERROR: Please specify the number of tests for each message size as ");
      fprintf(stderr,"second program argument. It must be > 0\n\n");
    }
    exit(0);
  }	

  /* Determine largest message category to test and determine smax */
  cat_max = atoi(argv[2]);
  if ((cat_max < 0) || (cat_max > CATEGORIES-1))
  {
    if (MASTER)
      fprintf(stderr,"ERROR: Maximum message size is 2^n floats. Please choose 0 <= n <= %i.\n", 
        CATEGORIES-1);
    exit(0);
  }	

  howmanyfloats[0] = 1;	
  for (i=1; i < CATEGORIES; i++)
    howmanyfloats[i] = howmanyfloats[i-1] * 2;

  smax = howmanyfloats[cat_max];
	 	 
  if (MASTER)
  {
    fprintf(stdout, "===--- Measuring All-to-all communication throughput ---===\n");
    fprintf(stdout, "number of processes:     %d\n", nprocs);
    fprintf(stdout, "processors on each node: %d\n", procs_per_node);
    fprintf(stdout, "number of tests:         %d\n", tests);
    fprintf(stdout, "Largest message size:    %d floats  (%d bytes)\n",smax, smax*((int) sizeof(float)));
    fprintf(stdout, "timer resolution:        %f seconds\n\n", t_res );
  }
  MPI_Barrier(MPI_COMM_WORLD);	
  fprintf(stdout, "procid=%d, hostname=%s, procs per node=%d\n",procid,mpi_hostname,procs_per_node);
  MPI_Barrier(MPI_COMM_WORLD);

  /* allocate some space for saving the timing results: */
  for (i = 0; i < CATEGORIES; i++)
  {    
    if ( (t_comm[i] = calloc(tests, sizeof(double))) == NULL )
    {
      fprintf(stderr, "ERROR: could not allocate memory for timing data. %d\n",procid);
      exit(0);
    }     
  }
	
  /* allocate space for send/receive data */
  if ( (sdata = calloc(nprocs,smax*sizeof(float))) == NULL )
  {
    fprintf(stderr,"ERROR: Could not allocate enough memory for send data. Hostname=%s, rank=%d, size=%d\n",
      mpi_hostname,procid,nprocs*smax*((int) sizeof(float)));
    exit(0);
  }
  if ( (rdata = calloc(nprocs,smax*sizeof(float))) == NULL )
  {
    fprintf(stderr,"ERROR: Could not allocate enough memory for receive data. Hostname=%s, rank=%d, size=%d\n",
      mpi_hostname,procid,nprocs*smax*((int) sizeof(float)));
    exit(0);
  }
								
#ifdef USE_MPE
  /* MPE_Init_log(); */
  /* Describe MPE states */
  for (i=0; i < CATEGORIES; i++)
  {
    ev_start[i]  = MPE_Log_get_event_number();
    ev_finish[i] = MPE_Log_get_event_number();
    sprintf(desc, "Cat %8d", i);
    MPE_Describe_state(ev_start[i], ev_finish[i], desc, color[i]);
  }
#endif


  /* --------- loop over communication schemes ------------------ */
  for (commtype = 0; commtype < 3; commtype++) 
  { 
    /* Do not test RSR scheme on nodes with more than one CPU! */
    if ((procs_per_node > 1) && (commtype == 1))
      commtype++;
		
    if (MASTER)
      fprintf(stdout, "\n--- Testing %s communication scheme: --- \n", typestring[commtype]);
    /* Do the tests! */
    for (i=OFFSET; i <= cat_max; i++)
    {
      t_aver[i] = 0.0;
      for (j=0; j<tests; j++)
      {
	if (MASTER)
	  fprintf(stdout, "%s (%d/%d) ", typestring[commtype],j+1,tests);		
        t_comm[i][j] = do_test(howmanyfloats[i],ev_start[i],ev_finish[i],commtype,
                               nprocs,procs_per_node,nnodes,procid,
                               sdata,rdata);
        t_aver[i] += t_comm[i][j];
        if (j==0)
        {
          t_min[i] = t_comm[i][j];
          t_max[i] = t_comm[i][j];
        }
        else
        {
          /* minimum time of a single test */
          if (t_comm[i][j] < t_min[i])
            t_min[i] = t_comm[i][j];
          /* maximum time of a single test */
          if (t_comm[i][j] > t_max[i])
            t_max[i] = t_comm[i][j];
        }
      }
      t_aver[i] /= tests;
    }
  
    /* Master outputs summary to screen and to file: */
    if (MASTER)
    {
      fprintf(stdout, "Summary for %s communication scheme (%d-run average, timer resolution %f):\n",
        typestring[commtype],tests,t_res);
      sprintf(filename,"%s-%dprocs-%dnodes-%d",typestring[commtype],nprocs,nnodes,tests);			  
      f_ptr=(FILE *)fopen(filename,"w");
      fprintf(f_ptr,"Timings in seconds for the %s communication scheme.\n", typestring[commtype]);
      fprintf(f_ptr,"CPUs: %d, CPUs per node: %d, tests done for each row: %d\n",
        nprocs, procs_per_node, tests);		
      fprintf(f_ptr,"    floats       bytes      t_av   std dev       t_min     t_max\n");			
      for (i = OFFSET; i <= cat_max; i++)
      {
        /* calculate stdandard deviation for each category: */
        dummy=0.0;
        for (j=0; j<tests; j++)
        {
          dum1   = t_comm[i][j] - t_aver[i];
          dummy += dum1*dum1;
        }
        dummy /= tests-1;
        sdev[i] = sqrt(dummy);
        fprintf(stdout, "%10d floats took %f (%f) seconds. Min: %f  max: %f\n",
          howmanyfloats[i],t_aver[i],sdev[i],t_min[i],t_max[i]);
        fprintf(f_ptr,"%10d  %10d  %f  %f    %f  %f\n",
          howmanyfloats[i],howmanyfloats[i]*((int) sizeof(float)),
          t_aver[i],sdev[i],t_min[i],t_max[i]);
      }
      fclose(f_ptr);
    }
  }
  /* ----(end) loop over communication schemes (end) ------------ */
	
  MPI_Barrier(MPI_COMM_WORLD);	
  MPI_Finalize();
  return 0;
} 
